

#include "quakedef.h"
#include "menu_common.h"
#include "winquake.h"

void Draw_Mapshot (char *name, int x, int y);
extern char *ID1Maps[];
extern char *ID1BSPs[];
extern char *HipnoMaps[];
extern char *HipnoBSPs[];
extern char *RogueMaps[];
extern char *RogueBSPs[];
extern char *SkillNames[];

CQMenu menu_Search (&menu_TCPIPJoinGame, m_other);
CQMenu menu_SList (&menu_TCPIPJoinGame, m_other);

/*
========================================================================================================================

					MULTI PLAYER MENU

========================================================================================================================
*/

int Menu_MultiplayerCustomDraw (int y)
{
	if (serialAvailable || ipxAvailable || tcpipAvailable) return y;

	Menu_PrintCenterWhite (vid.height - 80, "No Communications Available");

	return y;
}


// dummy cvars to make textboxes not overwrite the real ones
cvar_t dummy_name;
cvar_t dummy_hostname;

// player colours
int setup_shirt, setup_oldshirt;
int setup_pants, setup_oldpants;

void Menu_SetupApplyFunc (void)
{
	// execute the commands
	Cbuf_AddText (va ("name \"%s\"\n", dummy_name.string));
	Cvar_Set ("hostname", dummy_hostname.string);

	if (setup_shirt != setup_oldshirt || setup_pants != setup_oldpants)
		Cbuf_AddText (va ("color %i %i\n", setup_shirt, setup_pants));

	// return to the multiplayer menu
	menu_Multiplayer.EnterMenu ();
}


void Menu_SetupCustomEnter (void)
{
	// copy cvars out
	strncpy (dummy_name.string, cl_name.string, 1023);
	strncpy (dummy_hostname.string, hostname.string, 1023);

	setup_shirt = setup_oldshirt = ((int) cl_color.value) >> 4;
	setup_pants = setup_oldpants = ((int) cl_color.value) & 15;
}


byte identityTable[256];
byte translationTable[256];

void M_BuildTranslationTable (int top, int bottom)
{
	int		j;
	byte	*dest, *source;

	for (j = 0; j < 256; j++)
		identityTable[j] = j;

	dest = translationTable;
	source = identityTable;
	memcpy (dest, source, 256);

	if (top < 128)	// the artists made some backwards ranges.  sigh.
		memcpy (dest + TOP_RANGE, source + top, 16);
	else
		for (j = 0; j < 16; j++)
			dest[TOP_RANGE + j] = source[top + 15 - j];

	if (bottom < 128)
		memcpy (dest + BOTTOM_RANGE, source + bottom, 16);
	else
		for (j = 0; j < 16; j++)
			dest[BOTTOM_RANGE + j] = source[bottom + 15 - j];
}


int Menu_SetupCustomDraw (int y)
{
	qpic_t *p;
	int newy = y;

	p = Draw_CachePic ("gfx/bigbox.lmp");
	Draw_Pic ((vid.width - p->width) / 2, y, p);
	newy += p->height + 5;

	p = Draw_CachePic ("gfx/menuplyr.lmp");
	M_BuildTranslationTable (setup_shirt * 16, setup_pants * 16);
	Draw_PicTranslate ((vid.width - p->width) / 2, y + 8, p, translationTable);

	return newy;
}


int Menu_TCPIPCustomDraw (int y)
{
	Menu_Print (148 - strlen ("IP Address") * 8, y, "IP Address");
	Menu_PrintWhite (172, y, my_tcpip_address);

	y += 15;

	Menu_Print (148 - strlen ("Host Name") * 8, y, "Host Name");
	Menu_PrintWhite (172, y, hostname.string);

	y += 15;

	if (serialAvailable || ipxAvailable || tcpipAvailable) return y;

	Menu_PrintCenterWhite (vid.height - 80, "No Communications Available");

	return y;
}


cvar_t dummy_port;
cvar_t dummy_remoteip;

void Menu_TCPIPCustomEnter (void)
{
	// copy out the port
	sprintf (dummy_port.string, "%i", DEFAULTnet_hostport);
	dummy_remoteip.string[0] = 0;
}


void Menu_TCPIPPostConfigCommon (void)
{
	// stop all demos
	Cbuf_AddText ("stopdemo\n");

	// retrieve the port
	int l =  Q_atoi (dummy_port.string);

	// store it out
	if (l > 65535)
		net_hostport = DEFAULTnet_hostport;
	else net_hostport = l;
}


void Menu_TCPIPContinueToGameOptions (void)
{
	// complete net config
	Menu_TCPIPPostConfigCommon ();

	// fire the game options menu entry function
	menu_GameConfig.EnterMenu ();
}


cvar_t dummy_maxplayers;
cvar_t dummy_timelimit;
cvar_t dummy_fraglimit;
int dummy_skill;
int dummy_teamplay;

int map_number = 0;
char **MapNames = ID1Maps;
char **MapBSPs = ID1BSPs;

char *ID1Teamplay[] = {"Off", "No Friendly Fire", "Friendly Fire", NULL};
char *RogueTeamplay[] = {"Off", "No Friendly Fire", "Friendly Fire", "Tag", "Capture the Flag", "One Flag CTF", "Three Team CTF", NULL};
char **ActiveTeamplay = ID1Teamplay;
char *GameTypes[] = {"Cooperative", "Deathmatch", NULL};
int gametype_number = 0;


int Menu_GameConfigCustomDraw (int y)
{
	if ((int) dummy_maxplayers.value == svs.maxclientslimit)
	{
		if (svs.maxclientslimit == MAX_SCOREBOARD)
			Menu_PrintCenter (vid.height - 75, va ("Cannot select more than %i players", svs.maxclientslimit));
		else
		{
			Menu_PrintCenter (vid.height - 75, va ("Running with more than %i players", svs.maxclientslimit));
			Menu_PrintCenter (vid.height - 65, "Requires using command-line parameters");
			Menu_PrintCenter (vid.height - 50, "Please see TECHINFO.TXT");
		}
	}

	return y;
}


void Menu_GameConfigCustomEnter (void)
{
	// copy out cvars to prevent in-place editing of values screwing with current settings before we commit them
	// especially as many of these are broadcast cvars...
	dummy_maxplayers.value = svs.maxclients;
	dummy_fraglimit.value = fraglimit.value;
	dummy_timelimit.value = timelimit.value;
	dummy_skill = (int) (skill.value + 0.5);
	dummy_teamplay = (int) teamplay.value;

	// bound values so we don't crash or otherwise behave unexpectedly
	if (dummy_maxplayers.value < 2) dummy_maxplayers.value = 2; else if (dummy_maxplayers.value > svs.maxclientslimit) dummy_maxplayers.value = svs.maxclientslimit;
	if (dummy_fraglimit.value < 0) dummy_fraglimit.value = 0; else if (dummy_fraglimit.value > 100) dummy_fraglimit.value = 100;
	if (dummy_timelimit.value < 0) dummy_timelimit.value = 0; else if (dummy_timelimit.value > 60) dummy_timelimit.value = 60;
	if (dummy_skill < 0) dummy_skill = 0; else if (dummy_skill > 3) dummy_skill = 3;

	if (rogue) {if (dummy_teamplay < 0) dummy_teamplay = 0; else if (dummy_teamplay > 6) dummy_teamplay = 6;}
	else {if (dummy_teamplay < 0) dummy_teamplay = 0; else if (dummy_teamplay > 2) dummy_teamplay = 2;}

	if (coop.value)
		gametype_number = 0;
	else gametype_number = 1;
}


void Menu_GameConfigBeginGame (void)
{
	// disconnet any current servers
	if (sv.active) Cbuf_AddText ("disconnect\n");

	// so host_netport will be re-examined
	Cbuf_AddText ("listen 0\n");
	Cbuf_AddText (va ("maxplayers %u\n", (int) dummy_maxplayers.value));
	SCR_BeginLoadingPlaque ();

	// copy cvar values back now that we're committed
	Cvar_Set (&timelimit, dummy_timelimit.value);
	Cvar_Set (&fraglimit, dummy_fraglimit.value);
	Cvar_Set (&skill, (float) dummy_skill);
	Cvar_Set (&teamplay, (float) dummy_teamplay);

	// set gametype
	if (gametype_number)
	{
		Cvar_Set (&coop, 0.0f);
		Cvar_Set (&deathmatch, 1.0f);
	}
	else
	{
		Cvar_Set (&coop, 1.0f);
		Cvar_Set (&deathmatch, 0.0f);
	}

	// load the correct map
	Cbuf_AddText (va ("map %s\n", MapBSPs[map_number]));
}


bool Menu_SearchComplete = false;
double Menu_SearchCompleteTime;

void Menu_SearchCustomEnter (void)
{
	slistSilent = true;
	slistLocal = false;
	Menu_SearchComplete = false;
	NET_Slist_f ();
}


int Menu_SearchCustomDraw (int y)
{
	if (slistInProgress)
	{
		Menu_PrintCenterWhite (y + 50, "...Searching...");
		NET_Poll ();
		return y;
	}

	if (!Menu_SearchComplete)
	{
		Menu_SearchComplete = true;
		Menu_SearchCompleteTime = realtime;
	}

	if (hostCacheCount)
	{
		menu_SList.EnterMenu ();
		return y;
	}

	Menu_PrintCenterWhite (y + 50, "No Quake Servers found");

	if ((realtime - Menu_SearchCompleteTime) < 3.0) return y;

	menu_TCPIPJoinGame.EnterMenu ();

	return y;
}


CScrollBoxProvider *MenuSListScrollbox = NULL;

void Menu_SListOnDraw (int y, int itemnum)
{
	// draw the item
	Menu_Print (-8, y, hostcache[itemnum].name);
}


void Menu_SListOnHover (int initialy, int y, int itemnum)
{
	// highlight bar
	Menu_HighlightBar (-174, y, 172);

	hostcache_t *hc = &hostcache[itemnum];

	// rest of info
	Draw_Mapshot (va ("%s/maps/%s", com_gamedir, hc->map), (vid.width - 320) / 2 + 208, initialy + 8);
	Menu_PrintWhite (228, initialy + 145, "Server Info");
	Menu_Print (218, initialy + 160, DIVIDER_LINE);

	char hostipconfig[32];
	char *hostport;

	strcpy (hostipconfig, hc->cname);

	for (int i = strlen (hostipconfig) - 1; i; i--)
	{
		if (hostipconfig[i] == ':')
		{
			hostport = &hostipconfig[i + 1];
			hostipconfig[i] = 0;
			break;
		}
	}

	Menu_Print (212, initialy + 175, va ("IP Addr: %s", hostipconfig));
	Menu_Print (212, initialy + 185, va ("Port:    %s", hostport));
	Menu_Print (218, initialy + 200, DIVIDER_LINE);
	Menu_Print (212, initialy + 215, va ("Map:     %s", hc->map));
	Menu_Print (212, initialy + 225, va ("Players: %i of %i", hc->users, hc->maxusers));
}


char m_return_reason [32];
CQMenu *Menu_NetReturn = NULL;

float Net_ErrorReturnTime = 0;

void NET_MenuReturn (void)
{
	if (Menu_NetReturn)
	{
		Net_ErrorReturnTime = realtime + 5;
		Menu_NetReturn->EnterMenu ();
	}

	Menu_NetReturn = NULL;
}


int Menu_NetErrorCustomDraw (int y)
{
	if (Net_ErrorReturnTime < realtime) return y;

	Draw_TextBox ((vid.width / 2) - 175, 80, 334, 90);
	Menu_PrintCenterWhite (100, "Connection Error");
	Menu_PrintCenter (115, DIVIDER_LINE);
	Menu_PrintCenter (130, "Failed to connect to remote server");
	Menu_PrintCenter (145, va ("The reason given was:"));
	Menu_PrintCenter (160, va ("\"%s\"", m_return_reason));

	return y;
}


void Menu_MultiplayerConnectToGame (CQMenu *returnmenu, char *remote)
{
	key_dest = key_game;
	m_state = m_none;
	Menu_NetReturn = returnmenu;
	m_return_reason[0] = 0;
	Cbuf_AddText (va ("connect \"%s\"\n", remote));
}


void Menu_SListOnEnter (int itemnum)
{
	// join this game
	Menu_MultiplayerConnectToGame (&menu_SList, hostcache[itemnum].cname);
}


void Menu_SListCustomKey (int key)
{
	if (!MenuSListScrollbox) return;

	switch (key)
	{
	case K_DOWNARROW:
	case K_UPARROW:
		menu_soundlevel = m_sound_nav;

	case K_ENTER:
		MenuSListScrollbox->KeyFunc (key);
		break;
	}
}


void Menu_SListCustomEnter (void)
{
	// sort the server list
	if (hostCacheCount > 1)
	{
		int	i, j;
		hostcache_t temp;

		for (i = 0; i < hostCacheCount; i++)
		{
			for (j = i + 1; j < hostCacheCount; j++)
			{
				if (strcmp (hostcache[j].name, hostcache[i].name) < 0)
				{
					Q_memcpy (&temp, &hostcache[j], sizeof (hostcache_t));
					Q_memcpy (&hostcache[j], &hostcache[i], sizeof (hostcache_t));
					Q_memcpy (&hostcache[i], &temp, sizeof (hostcache_t));
				}
			}
		}
	}

	// create a scrollbox for it
	if (MenuSListScrollbox) delete MenuSListScrollbox;
	MenuSListScrollbox = new CScrollBoxProvider (hostCacheCount, 20, 22);
	MenuSListScrollbox->SetDrawItemCallback (Menu_SListOnDraw);
	MenuSListScrollbox->SetHoverItemCallback (Menu_SListOnHover);
	MenuSListScrollbox->SetEnterItemCallback (Menu_SListOnEnter);
}


int Menu_SListCustomDraw (int y)
{
	if (!MenuSListScrollbox) return y;

	// draw through the scrollbox provider
	MenuSListScrollbox->DrawItems ((vid.width - 320) / 2 - 24, y);

	return y;
}


void Menu_TCPIPJoinGameCommand (void)
{
	// connect to this game
	Menu_MultiplayerConnectToGame (&menu_TCPIPJoinGame, dummy_remoteip.string);
}


void Menu_InitMultiplayerMenu (void)
{
	// top level menu
	menu_Multiplayer.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_Multiplayer.AddOption (new CQMenuTitle ("Configure Multiplayer Game Options"));
	menu_Multiplayer.AddOption (new CQMenuSubMenu ("Join an Existing Multiplayer Game", &menu_TCPIPJoinGame));
	menu_Multiplayer.AddOption (new CQMenuSpacer ());
	menu_Multiplayer.AddOption (new CQMenuSubMenu ("Start a New Multiplayer Game", &menu_TCPIPNewGame));
	menu_Multiplayer.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Multiplayer.AddOption (new CQMenuSubMenu ("Setup Player Options", &menu_Setup));
	menu_Multiplayer.AddOption (new CQMenuCustomDraw (Menu_MultiplayerCustomDraw));

	// setup menu
	menu_Setup.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_Setup.AddOption (new CQMenuTitle ("Setup Player Options"));
	menu_Setup.AddOption (new CQMenuCustomEnter (Menu_SetupCustomEnter));
	menu_Setup.AddOption (new CQMenuCvarTextbox ("Player Name", &dummy_name));
	menu_Setup.AddOption (new CQMenuCvarTextbox ("Host Name", &dummy_hostname));
	menu_Setup.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Setup.AddOption (new CQMenuColourBar ("Shirt Colour", &setup_shirt));
	menu_Setup.AddOption (new CQMenuCustomDraw (Menu_SetupCustomDraw));
	menu_Setup.AddOption (new CQMenuColourBar ("Pants Colour", &setup_pants));
	menu_Setup.AddApplyCommand ("Apply Changes", Menu_SetupApplyFunc);

	// start a new game
	menu_TCPIPNewGame.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_TCPIPNewGame.AddOption (new CQMenuTitle ("Configure TCP/IP Options - New Game"));
	menu_TCPIPNewGame.AddOption (new CQMenuCustomEnter (Menu_TCPIPCustomEnter));
	menu_TCPIPNewGame.AddOption (new CQMenuCustomDraw (Menu_TCPIPCustomDraw));
	menu_TCPIPNewGame.AddOption (new CQMenuCvarTextbox ("Port", &dummy_port, true));
	menu_TCPIPNewGame.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_TCPIPNewGame.AddOption (new CQMenuCommand ("Continue to Game Options", Menu_TCPIPContinueToGameOptions));

	// join an existing game
	menu_TCPIPJoinGame.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_TCPIPJoinGame.AddOption (new CQMenuTitle ("Configure TCP/IP Options - Join a Game"));
	menu_TCPIPJoinGame.AddOption (new CQMenuCustomEnter (Menu_TCPIPCustomEnter));
	menu_TCPIPJoinGame.AddOption (new CQMenuCustomDraw (Menu_TCPIPCustomDraw));
	menu_TCPIPJoinGame.AddOption (new CQMenuCvarTextbox ("Port", &dummy_port, true));
	menu_TCPIPJoinGame.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_TCPIPJoinGame.AddOption (new CQMenuSubMenu ("Search for Local Games", &menu_Search));
	menu_TCPIPJoinGame.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_TCPIPJoinGame.AddOption (new CQMenuCvarTextbox ("Remote Server", &dummy_remoteip));
	menu_TCPIPJoinGame.AddOption (new CQMenuCommand ("Join this Game", Menu_TCPIPJoinGameCommand));
	menu_TCPIPJoinGame.AddOption (new CQMenuCustomDraw (Menu_NetErrorCustomDraw));

	// game options
	if (rogue) ActiveTeamplay = RogueTeamplay;
	menu_GameConfig.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_GameConfig.AddOption (new CQMenuTitle ("Game Configuration Options"));
	menu_GameConfig.AddOption (new CQMenuCustomEnter (Menu_GameConfigCustomEnter));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Game Type", &gametype_number, GameTypes));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Teamplay", &dummy_teamplay, ActiveTeamplay));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Maximum Players", &dummy_maxplayers, 2, svs.maxclientslimit, 1, NULL, NULL));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Skill", &dummy_skill, SkillNames));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Frag Limit", &dummy_fraglimit, 0, 100, 10, "None", "Frags"));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Time Limit", &dummy_timelimit, 0, 60, 5, "None", "Minutes"));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuSpacer ("Select a Map"));
	menu_GameConfig.AddOption (new CQMenuCustomDraw (Menu_GameConfigCustomDraw));

	// set up the correct maps menu
	if (hipnotic)
	{
		MapNames = HipnoMaps;
		MapBSPs = HipnoBSPs;
	}
	else if (rogue)
	{
		MapNames = RogueMaps;
		MapBSPs = RogueBSPs;
	}

	// and add it
	menu_GameConfig.AddOption (new CQMenuSpinControl (NULL, &map_number, MapNames));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuCommand ("Begin New Multiplayer Game", Menu_GameConfigBeginGame));

	// search for games
	menu_Search.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_Search.AddOption (new CQMenuTitle ("Search for Local Games"));
	menu_Search.AddOption (new CQMenuCustomEnter (Menu_SearchCustomEnter));
	menu_Search.AddOption (new CQMenuCustomDraw (Menu_SearchCustomDraw));

	// server list
	menu_SList.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_SList.AddOption (new CQMenuTitle ("Local Quake Server List"));
	menu_SList.AddOption (new CQMenuCustomEnter (Menu_SListCustomEnter));
	menu_SList.AddOption (new CQMenuCustomDraw (Menu_SListCustomDraw));
	menu_SList.AddOption (new CQMenuCustomDraw (Menu_NetErrorCustomDraw));
	menu_SList.AddOption (new CQMenuCustomKey (K_UPARROW, Menu_SListCustomKey));
	menu_SList.AddOption (new CQMenuCustomKey (K_DOWNARROW, Menu_SListCustomKey));
	menu_SList.AddOption (new CQMenuCustomKey (K_ENTER, Menu_SListCustomKey));
}


